'' =================================================================================================
''
''   File....... jm_servo8_ez.spin
''   Purpose.... Servo control without PASM!
''   Author..... Jon "JonnyMac" McPhalen (aka Jon Williams)
''               Copyright (c) 2011 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 19 JUL 2011
''
'' =================================================================================================


con

  SVO_MIN = 0_600                                               ' for 180 degree servos
  SVO_MAX = 2_400


var

  long  us001                                                   ' ticks per us

  long  smin                                                    ' min position
  long  smax                                                    ' max position

  long  pos[8]                                                  ' current position (usecs)
  long  target[8]                                               ' target position (usecs)
  long  delta[8]                                                ' delta/cycle (usecs)

  long  cog                                                     ' cog # of servo driver
  long  stack[16]                                               ' stack for Spin cog
 
  
pub start(count, base)

'' Start servo cog
'' -- count is number of servo outputs (up to eight)
'' -- base is first pin of contiguous group

  return startx(count, base, SVO_MIN, SVO_MAX)                  ' start with standard range


pub startx(count, base, posmin, posmax) : ok

  us001 := clkfreq / 1_000_000                                  ' ticks per microsecond

  smin := posmin #> SVO_MIN                                     ' set range
  smax := posmax <# SVO_MAX

  stop                                                          ' stop if already running
  ok := cog := cognew(servo8(count, base), @stack) + 1          ' launch servo cog

  return ok
    

pub stop

'' Stop servo cog if running

  if cog
    cogstop(cog~ - 1)                                           ' stop the cog

  set_all(1_500, 0)


pub set_pos(ch, newpos, speed)

'' Set selected servo (ch) to new position
'' -- newpos specified in microseconds
'' -- speed specified in us change per second

  if (ch => 0) and (ch =< 7)                                    ' legal channel?
    newpos := smin #> newpos <# smax                            ' keep legal
    target[ch] := newpos                                        ' set new target
    if (speed =< 0)                                             ' check speed
      delta[ch] := 0                                            ' if 0
      pos[ch] := newpos                                         '  move now
    else
      ' round up
      delta[ch] := ((speed * 10 / 50) + 5) / 10 #> 1            ' divide by 50/updates per second 


pub set_all(newpos, speed)

'' Set all servos to new position
'' -- newpos specified in microseconds
'' -- speed specified in us change per second

  newpos := smin #> newpos <# smax                              ' keep legal
  speed  := 0 #> speed

  longfill(@target, newpos, 8)                                  ' update all positions 
  if (speed == 0)                                               ' if 0
    longfill(@pos, newpos, 8)                                   '   move now
  else
    ' round up
    speed := ((speed * 10 / 50) + 5) / 10 #> 1                  ' divide by 50/updates per second
  longfill(@delta, speed, 8)                                    ' update speed settings   


pub move_to(ch, newpos, ms) | frames, move

'' Moves select servo to newpos in ~ms milliseconds
'' -- ms resolution is 20 (servo frame timing)
'' -- not accurate at low ms values for long moves

  if (ch => 0) and (ch =< 7)                                    ' legal channel?
    newpos := smin #> newpos <# smax                            ' keep legal
    frames := ms / 20                                           ' calc frames in move
    move := ||(newpos - pos[ch]) * 10 / frames                  ' calc 0.1us/frame
    move := ((move + 5) / 10) #> 1                              ' round up to 1us/frame
    if (frames < 2)
      pos[ch] := newpos                                         '   move now
      target[ch] := newpos
    else
      target[ch] := newpos                                      '   set new target            
      delta[ch] := move                                         '   and movement delta/frame  


pub move_all(newpos, ms) | idx

'' Moves all servos to newpos in ~ms milliseconds
'' -- ms resolution is 20 (servo frame timing)
'' -- not accurate at low ms values for long moves

  repeat idx from 0 to 7
    move_to(idx, newpos, ms)

  
pub wait(ch)

'' Waits (blocks program) until servo channel reaches target position

  repeat while (pos[ch] <> target[ch])


pri servo8(count, base) | slot, idx                             ' launch with cognew

'' Runs up to 8 servos 

  count := 1 #> count <# 8                                      ' keep legal

  outa := 0                                                     ' all off
  dira := ($FFFF_FFFF >> (32 - count)) << base                  ' set outputs

  frqa := 1                                                     ' preset for counter
  phsa := 0 

  slot := cnt                                                   ' start slot timing
  repeat
    repeat idx from 0 to 7                                      ' loop through servos
      if (idx < count)                                          ' active channel?
        ctra := (%00100 << 26 ) | (base + idx)                  ' PWM/NCO mode on servo pin
        phsa := -(pos[idx] * us001)                             ' set pulse timing
        if (pos[idx] < target[idx])                             ' update for speed
          pos[idx] := (pos[idx] + delta[idx]) <# target[idx]
        elseif (pos[idx] > target[idx])
          pos[idx] := (pos[idx] - delta[idx]) #> target[idx]
      waitcnt(slot += (2_500 * us001))                          ' let slot finish
      ctra := 0                                                 ' release ctra from pin 


dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}